"""Decode PaVE video packets from the AR.Drone into RGB image buffers."""

import struct

import av


# PaVE header format (64 bytes, little-endian packed)
# See https://developer.parrot.com/docs/SDK3/ for protocol details
_PAVE_HEADER_FMT = "<4sBBHI HH HH IIBBBBIIHBBBBxx I 12x"
_PAVE_HEADER_SIZE = struct.calcsize(_PAVE_HEADER_FMT)

# Field indices in the unpacked header tuple
_F_SIGNATURE = 0
_F_HEADER_SIZE = 2
_F_PAYLOAD_SIZE = 3


class DecodeError(Exception):
    """Raised when a video packet cannot be decoded."""


class Decoder:
    """Persistent H.264 decoder for PaVE video streams.

    Maintains codec state across calls so that P-frames can reference
    previous I-frames, matching the behavior of the original C extension.
    """

    def __init__(self):
        self._codec = av.CodecContext.create("h264", "r")

    def decode(self, data):
        """Decode a PaVE packet into an (width, height, rgb_bytes) tuple.

        Args:
            data: Raw bytes of a complete PaVE packet (header + payload).

        Returns:
            Tuple of (width, height, bytes) where bytes is RGB24 image data.

        Raises:
            DecodeError: If the packet signature is invalid, the size is
                inconsistent, or the H.264 payload cannot be decoded.
        """
        if len(data) < _PAVE_HEADER_SIZE:
            raise DecodeError("packet too short for PaVE header")

        fields = struct.unpack_from(_PAVE_HEADER_FMT, data)

        if fields[_F_SIGNATURE] != b"PaVE":
            raise DecodeError("packet did not have correct signature")

        header_size = fields[_F_HEADER_SIZE]
        payload_size = fields[_F_PAYLOAD_SIZE]

        if header_size + payload_size != len(data):
            raise DecodeError(
                "packet size did not match expected size from header"
            )

        payload = data[header_size : header_size + payload_size]

        packet = av.Packet(payload)
        try:
            frames = self._codec.decode(packet)
        except av.error.InvalidDataError as exc:
            raise DecodeError("could not decode frame") from exc

        if not frames:
            raise DecodeError("could not decode frame")

        frame = frames[0]
        rgb_frame = frame.to_rgb()
        image = bytes(rgb_frame.planes[0])
        return rgb_frame.width, rgb_frame.height, image


# Module-level decoder instance, preserving the original C extension's
# behavior of maintaining codec state across calls.
_decoder = Decoder()


def decode(data):
    """Decode a PaVE packet into an (width, height, rgb_bytes) tuple.

    This is the module-level convenience function matching the original
    C extension's ``ardrone.video.decode()`` interface.
    """
    return _decoder.decode(data)



import struct


def decode(packet):
    """Decode a navdata packet."""
    offset = 0

    _ = struct.unpack_from('IIII', packet, offset)
    s = _[1]
    state = dict()
    state['fly']                 = s       & 1  # FLY MASK : (0) ardrone is landed, (1) ardrone is flying
    state['video']               = s >>  1 & 1  # VIDEO MASK : (0) video disable, (1) video enable
    state['vision']              = s >>  2 & 1  # VISION MASK : (0) vision disable, (1) vision enable
    state['control']             = s >>  3 & 1  # CONTROL ALGO (0) euler angles control, (1) angular speed control
    state['altitude']            = s >>  4 & 1  # ALTITUDE CONTROL ALGO : (0) altitude control inactive (1) altitude control active
    state['user_feedback_start'] = s >>  5 & 1  # USER feedback : Start button state
    state['command']             = s >>  6 & 1  # Control command ACK : (0) None, (1) one received
    state['fw_file']             = s >>  7 & 1  # Firmware file is good (1)
    state['fw_ver']              = s >>  8 & 1  # Firmware update is newer (1)
    state['fw_upd']              = s >>  9 & 1  # Firmware update is ongoing (1)
    state['navdata_demo']        = s >> 10 & 1  # Navdata demo : (0) All navdata, (1) only navdata demo
    state['navdata_bootstrap']   = s >> 11 & 1  # Navdata bootstrap : (0) options sent in all or demo mode, (1) no navdata options sent
    state['motors']              = s >> 12 & 1  # Motor status : (0) Ok, (1) Motors problem
    state['com_lost']            = s >> 13 & 1  # Communication lost : (1) com problem, (0) Com is ok
    state['vbat_low']            = s >> 15 & 1  # VBat low : (1) too low, (0) Ok
    state['user_el']             = s >> 16 & 1  # User Emergency Landing : (1) User EL is ON, (0) User EL is OFF
    state['timer_elapsed']       = s >> 17 & 1  # Timer elapsed : (1) elapsed, (0) not elapsed
    state['angles_out_of_range'] = s >> 19 & 1  # Angles : (0) Ok, (1) out of range
    state['ultrasound']          = s >> 21 & 1  # Ultrasonic sensor : (0) Ok, (1) deaf
    state['cutout']              = s >> 22 & 1  # Cutout system detection : (0) Not detected, (1) detected
    state['pic_version']         = s >> 23 & 1  # PIC Version number OK : (0) a bad version number, (1) version number is OK
    state['atcodec_thread_on']   = s >> 24 & 1  # ATCodec thread ON : (0) thread OFF (1) thread ON
    state['navdata_thread_on']   = s >> 25 & 1  # Navdata thread ON : (0) thread OFF (1) thread ON
    state['video_thread_on']     = s >> 26 & 1  # Video thread ON : (0) thread OFF (1) thread ON
    state['acq_thread_on']       = s >> 27 & 1  # Acquisition thread ON : (0) thread OFF (1) thread ON
    state['ctrl_watchdog']       = s >> 28 & 1  # CTRL watchdog : (1) delay in control execution (> 5ms), (0) control is well scheduled
    state['adc_watchdog']        = s >> 29 & 1  # ADC Watchdog : (1) delay in uart2 dsr (> 5ms), (0) uart2 is good
    state['com_watchdog']        = s >> 30 & 1  # Communication Watchdog : (1) com problem, (0) Com is ok
    state['emergency']           = s >> 31 & 1  # Emergency landing : (0) no emergency, (1) emergency

    data = dict()
    data['state'] = state
    data['header'] = _[0]
    data['sequence'] = _[2]
    data['vision'] = _[3]

    offset += struct.calcsize('IIII')

    demo_fields = [
        'ctrl_state',
        'battery',
        'theta',
        'phi',
        'psi',
        'altitude',
        'vx',
        'vy',
        'vz',
        'num_frames'
    ]
    angles = ['theta', 'phi', 'psi']
    while True:
        try:
            id_nr, size = struct.unpack_from('HH', packet, offset)
            offset += struct.calcsize('HH')
        except struct.error:
            break

        values = []
        for i in range(size - struct.calcsize('HH')):
            values.append(struct.unpack_from('c', packet, offset)[0])
            offset += struct.calcsize('c')

        if id_nr == 0:
            values = struct.unpack_from('IIfffIfffI', b''.join(values))
            demo = dict(zip(demo_fields, values))
            for a in angles:
                demo[a] = int(demo[a] / 1000)

            data['demo'] = demo

    return data



import socket
import struct
import threading

import ardrone.constant


def f2i(f):
    """Interpret IEEE-754 floating-point value as signed integer.

    Arguments:
    f -- floating point value
    """
    return struct.unpack('i', struct.pack('f', f))[0]


class ATCommand(object):
    def __init__(self, host):
        """
        Open a new AT command socket

        Parameters:
        host -- destination address
        """
        self.host = host
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.seq = 1
        self.interval = 0.2

        self.lock = threading.Lock()
        self.comwdg_timer = threading.Timer(self.interval, self.comwdg)

    def halt(self):
        """
        Halts communication with the drone
        """
        with self.lock:
            self.comwdg_timer.cancel()

    def ref(self, takeoff, emergency=False):
        """
        Basic behaviour of the drone: take-off/landing, emergency stop/reset)

        Parameters:
        takeoff -- True: Takeoff / False: Land
        emergency -- True: Turn off the engines
        """
        p = 0b10001010101000000000000000000
        if takeoff:
            p |= 0b1000000000
        if emergency:
            p |= 0b100000000
        self.at('REF', [p])

    def pcmd(self, progressive, lr, fb, vv, va):
        """
        Makes the drone move (translate/rotate).

        Parameters:
        progressive -- True: enable progressive commands, False: disable (i.e.
            enable hovering mode)
        lr -- left-right tilt: float [-1..1] negative: left, positive: right
        rb -- front-back tilt: float [-1..1] negative: forwards, positive:
            backwards
        vv -- vertical speed: float [-1..1] negative: go down, positive: rise
        va -- angular speed: float [-1..1] negative: spin left, positive: spin
            right

        The above float values are a percentage of the maximum speed.
        """
        p = 1 if progressive else 0
        self.at('PCMD', [p, float(lr), float(fb), float(vv), float(va)])

    def ftrim(self):
        """
        Tell the drone it's lying horizontally.
        """
        self.at('FTRIM')

    def zap(self, stream):
        """
        Selects which video stream to send on the video UDP port.

        Parameters:
        stream -- Integer: video stream to broadcast
        """
        # FIXME: improve parameters to select the modes directly
        self.at('ZAP', [stream])

    def config(self, option, value):
        """Set configuration parameters of the drone."""
        self.at('CONFIG', [str(option), str(value)])

    def comwdg(self):
        """
        Reset communication watchdog.
        """
        self.at('COMWDG')

    def aflight(self, flag):
        """
        Makes the drone fly autonomously.

        Parameters:
        flag -- Integer: 1: start flight, 0: stop flight
        """
        self.at('AFLIGHT', [flag])

    def pwm(self, m1, m2, m3, m4):
        """
        Sends control values directly to the engines, overriding control loops.

        Parameters:
        m1 -- Integer: front left command
        m2 -- Integer: front right command
        m3 -- Integer: back right command
        m4 -- Integer: back left command
        """
        self.at('PWM', [m1, m2, m3, m4])

    def led(self, anim, f, d):
        """
        Control the drones LED.

        Parameters:
        anim -- Integer: animation to play
        f -- Float: frequency in HZ of the animation
        d -- Integer: total duration in seconds of the animation
        """
        self.at('LED', [anim, float(f), d])

    def anim(self, anim, d):
        """
        Makes the drone execute a predefined movement (animation).

        Parameters:
        anim -- Integer: animation to play
        d -- Integer: total duration in seconds of the animation
        """
        self.at('ANIM', [anim, d])

    def at(self, command, params=[]):
        """
        Encodes and sends AT command

        Parameters:
        command -- the command
        params -- a list of elements which can be either int, float or string
        """
        params_str = []
        for p in params:
            if type(p) == int:
                params_str.append('{:d}'.format(p))
            elif type(p) == float:
                params_str.append('{:d}'.format(f2i(p)))
            elif type(p) == str:
                params_str.append('"{:s}"'.format(p))

        with self.lock:
            self.comwdg_timer.cancel()

            msg = 'AT*{:s}={:d}{:s}\r'.format(command, self.seq, ''.join(',' + param for param in params_str))
            self.sock.sendto(msg.encode(), (self.host, ardrone.constant.COMMAND_PORT))

            self.seq += 1

            self.comwdg_timer = threading.Timer(self.interval, self.comwdg)
            self.comwdg_timer.start()



"""
Python library for the AR.Drone.
"""

import time
import multiprocessing

import PIL.Image

import ardrone.at
import ardrone.network


class ARDrone(object):
    """ARDrone Class.

    Instantiate this class to control your drone and receive decoded video and
    navdata.
    """

    def __init__(self, host='192.168.1.1'):
        self.host = host

        self.speed = 0.2

        self.atcmd = ardrone.at.ATCommand(self.host)
        self.atcmd.config('general:navdata_demo', 'TRUE')
        self.atcmd.config('control:altitude_max', '20000')
        self.video_pipe, video_pipe_other = multiprocessing.Pipe()
        self.nav_pipe, nav_pipe_other = multiprocessing.Pipe()
        self.com_pipe, com_pipe_other = multiprocessing.Pipe()
        self.network_process = ardrone.network.ARDroneNetworkProcess(self.host, nav_pipe_other, video_pipe_other, com_pipe_other)
        self.network_process.start()
        self.ipc_thread = ardrone.network.IPCThread(self)
        self.ipc_thread.start()

        self.image = PIL.Image.new('RGB', (640, 360))
        self.navdata = dict()

        self.time = 0

    def takeoff(self):
        """Make the drone takeoff."""
        self.atcmd.ref(True)

    def land(self):
        """Make the drone land."""
        self.atcmd.ref(False)

    def hover(self):
        """Make the drone hover."""
        self.atcmd.pcmd(False, 0, 0, 0, 0)

    def move_left(self):
        """Make the drone move left."""
        self.atcmd.pcmd(True, -self.speed, 0, 0, 0)

    def move_right(self):
        """Make the drone move right."""
        self.atcmd.pcmd(True, self.speed, 0, 0, 0)

    def move_up(self):
        """Make the drone rise upwards."""
        self.atcmd.pcmd(True, 0, 0, self.speed, 0)

    def move_down(self):
        """Make the drone decent downwards."""
        self.atcmd.pcmd(True, 0, 0, -self.speed, 0)

    def move_forward(self):
        """Make the drone move forward."""
        self.atcmd.pcmd(True, 0, -self.speed, 0, 0)

    def move_backward(self):
        """Make the drone move backwards."""
        self.atcmd.pcmd(True, 0, self.speed, 0, 0)

    def turn_left(self):
        """Make the drone rotate left."""
        self.atcmd.pcmd(True, 0, 0, 0, -self.speed)

    def turn_right(self):
        """Make the drone rotate right."""
        self.atcmd.pcmd(True, 0, 0, 0, self.speed)

    def reset(self):
        """Toggle the drone's emergency state."""
        self.atcmd.ref(False, True)
        time.sleep(0.1)
        self.atcmd.ref(False, False)

    def trim(self):
        """Flat trim the drone."""
        self.atcmd.ftrim

    def set_cam(self, cam):
        """Set active camera.

        Valid values are 0 for the front camera and 1 for the bottom camera
        """
        self.atcmd.config('video:video_channel', cam)

    def set_speed(self, speed):
        """Set the drone's speed.

        Valid values are floats from [0..1]
        """
        self.speed = speed

    def halt(self):
        """Shutdown the drone.

        This method does not land or halt the actual drone, but the
        communication with the drone. You should call it at the end of your
        application to close all sockets, pipes, processes and threads related
        with this object.
        """
        self.atcmd.halt()
        self.ipc_thread.stop()
        self.ipc_thread.join()
        self.network_process.terminate()
        self.network_process.join()

    def move(self, lr, fb, vv, va):
        """Makes the drone move (translate/rotate).

        Parameters:
        lr -- left-right tilt: float [-1..1] negative: left, positive: right
        fb -- front-back tilt: float [-1..1] negative: forwards, positive:
            backwards
        vv -- vertical speed: float [-1..1] negative: go down, positive: rise
        va -- angular speed: float [-1..1] negative: spin left, positive: spin
            right"""
        self.atcmd.pcmd(True, lr, fb, vv, va)



import fcntl
import sys
import termios
import os

import ardrone


def main():
    if len(sys.argv) != 1 and len(sys.argv) != 2:
        print('usage: ' + sys.argv[0] + ' [host]')
        sys.exit(1)

    fd = sys.stdin.fileno()

    oldterm = termios.tcgetattr(fd)
    newattr = termios.tcgetattr(fd)
    newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
    termios.tcsetattr(fd, termios.TCSANOW, newattr)

    oldflags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, oldflags | os.O_NONBLOCK)

    if len(sys.argv) >= 2:
        drone = ardrone.ARDrone(sys.argv[1])
    else:
        drone = ardrone.ARDrone()

    try:
        while 1:
            try:
                c = sys.stdin.read(1).lower()

                if c == 'p':
                    break
                elif c == 'a':
                    drone.move_left()
                elif c == 'd':
                    drone.move_right()
                elif c == 'w':
                    drone.move_forward()
                elif c == 's':
                    drone.move_backward()
                elif c == ' ':
                    drone.land()
                elif c == '\n':
                    drone.takeoff()
                elif c == 'q':
                    drone.turn_left()
                elif c == 'e':
                    drone.turn_right()
                elif c == '1':
                    drone.move_up()
                elif c == '2':
                    drone.hover()
                elif c == '3':
                    drone.move_down()
                elif c == 't':
                    drone.reset()
                elif c == 'x':
                    drone.hover()
                elif c == 'y':
                    drone.trim()
            except IOError:
                pass
    finally:
        termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
        fcntl.fcntl(fd, fcntl.F_SETFL, oldflags)
        drone.halt()


if __name__ == '__main__':
    main()



"""
This module provides access to the data provided by the AR.Drone.
"""

import select
import socket
import struct
import threading
import multiprocessing

import PIL.Image

import ardrone.constant
import ardrone.navdata
import ardrone.video


class ARDroneNetworkProcess(multiprocessing.Process):
    """ARDrone Network Process.

    This process collects data from the video and navdata port, converts the
    data and sends it to the IPCThread.
    """

    def __init__(self, host, nav_pipe, video_pipe, com_pipe):
        multiprocessing.Process.__init__(self)
        self.nav_pipe = nav_pipe
        self.video_pipe = video_pipe
        self.com_pipe = com_pipe
        self.host = host

    def run(self):
        video_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        video_socket.connect((self.host, ardrone.constant.VIDEO_PORT))

        nav_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        nav_socket.setblocking(False)
        nav_socket.sendto(b'\x01\x00\x00\x00', (self.host, ardrone.constant.NAVDATA_PORT))

        stopping = False
        while not stopping:
            inputready, outputready, exceptready = select.select([nav_socket, video_socket, self.com_pipe], [], [])
            for i in inputready:
                if i == video_socket:
                    # get first few bytes of header
                    data = video_socket.recv(12, socket.MSG_WAITALL)
                    if len(data) != 12:
                        continue

                    # decode relevant portions of the header
                    sig_p, sig_a, sig_v, sig_e, version, codec, header, payload = struct.unpack('4cBBHI', data)

                    # check signature (and ignore packet otherwise)
                    if sig_p != b'P' or sig_a != b'a' or sig_v != b'V' or sig_e != b'E':
                        continue

                    # get remaining frame
                    data += video_socket.recv(header - 12 + payload, socket.MSG_WAITALL)

                    try:
                        # decode the frame
                        image = ardrone.video.decode(data)
                        self.video_pipe.send(image)
                    except ardrone.video.DecodeError:
                        pass
                elif i == nav_socket:
                    while 1:
                        try:
                            data, _ = nav_socket.recvfrom(65535)
                        except IOError:
                            # we consumed every packet from the socket and
                            # continue with the last one
                            break
                    navdata = ardrone.navdata.decode(data)
                    self.nav_pipe.send(navdata)
                elif i == self.com_pipe:
                    _ = self.com_pipe.recv()
                    stopping = True
                    break
        video_socket.close()
        nav_socket.close()


class IPCThread(threading.Thread):
    """Inter Process Communication Thread.

    This thread collects the data from the ARDroneNetworkProcess and forwards
    it to the ARDrone.
    """

    def __init__(self, drone):
        threading.Thread.__init__(self)
        self.drone = drone
        self.stopping = False

    def run(self):
        while not self.stopping:
            inputready, outputready, exceptready = select.select([self.drone.video_pipe, self.drone.nav_pipe], [], [], 1)
            for i in inputready:
                if i == self.drone.video_pipe:
                    while self.drone.video_pipe.poll():
                        width, height, image = self.drone.video_pipe.recv()
                    self.drone.image = PIL.Image.frombuffer('RGB', (width, height), image, 'raw', 'RGB', 0, 1)
                elif i == self.drone.nav_pipe:
                    while self.drone.nav_pipe.poll():
                        navdata = self.drone.nav_pipe.recv()
                    self.drone.navdata = navdata

    def stop(self):
        """Stop the IPCThread activity."""
        self.stopping = True
